# Copyright (C) <2022>  <Zheng-Zhi Sun>
# This file is part of QuantumIntelligence.
# QuantumIntelligence is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Foobar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Foobar. If not, see <https://www.gnu.org/licenses/>.

import torch as tc
import numpy as np


pauli_x = tc.tensor([[0, 1], [1, 0]])
pauli_y = tc.tensor([[0, -1j], [1j, 0]])
pauli_z = tc.tensor([[1, 0], [0, -1]])


class Gate:
    def __init__(self, unitary, label=None, inverse=False, independent=True, check_unitary=False, unitary_acc=1e-6,
                 is_fast=False):
        if isinstance(unitary, Gate):
            if independent:
                self.tensor = unitary.tensor.clone().detach()
            else:
                self.tensor = unitary.tensor
            self.inverse = unitary.inverse
            self.label = unitary.label
            if inverse:
                self.inverse = not self.inverse
        elif isinstance(unitary, tc.Tensor):
            if independent:
                self.tensor = unitary.clone().detach()
            else:
                self.tensor = unitary
            self.inverse = inverse
            self.label = label
        if check_unitary:
            self.check_unitary(unitary_acc)
        self.is_fast = is_fast

    @property
    def requires_grad(self):
        return self.tensor.requires_grad

    @requires_grad.setter
    def requires_grad(self, requires_grad):
        self.tensor.requires_grad = requires_grad

    @property
    def grad(self):
        return self.tensor.grad

    @grad.setter
    def grad(self, value):
        self.tensor.grad = value

    # core functions start

    def check_unitary(self, acc=1e-6):
        identity_matrix = tc.eye(self.tensor.shape[0], device=self.tensor.device, dtype=self.tensor.dtype)
        diff = tc.dist(self.tensor.mm(self.tensor.conj().t()), identity_matrix)
        flag = (diff > acc)
        if flag:
            raise ValueError('the gate is not unitary. distance is ', diff)

    def inv(self):
        self.inverse = not self.inverse

    def square(self):
        if self.tensor.is_sparse:
            self.tensor = tc.sparse.mm(self.tensor, self.tensor)
        else:
            self.tensor = self.tensor.mm(self.tensor)

    def controlled_gate(self, n_control=1, output=False):
        # this function has not been tested
        c_gate = self.tensor.clone().detach()
        a_m = tc.tensor([[1, 0], [0, 0]], device=self.tensor.device, dtype=self.tensor.dtype)
        b_m = tc.tensor([[0, 0], [0, 1]], device=self.tensor.device, dtype=self.tensor.dtype)
        new_shape = c_gate.shape
        for nn in range(n_control):
            n_dim = round(c_gate.numel() ** 0.5)
            eye_m = tc.eye(n_dim, dtype=self.tensor.dtype, device=self.tensor.device)
            new_shape = new_shape + (2, 2)
            c_gate = tc.kron(a_m, eye_m) + tc.kron(b_m, c_gate.view(n_dim, n_dim))
        if output:
            return c_gate.view(new_shape)
        else:
            self.tensor = c_gate

    def to(self, device_or_dtype):
        self.tensor = self.tensor.to(device_or_dtype)
        return self

    def __copy__(self):
        return Gate(self)

    # core function end

# Experimental univerisal gate set


def rx_gate(theta=np.pi, device='cpu', dtype=tc.complex64):
    # exp(-i*theta * x/2)
    gate = tc.eye(2)*np.cos(theta/2) - 1j * np.sin(theta/2)*pauli_x
    return Gate(gate.to(device).to(dtype), label=('RX', theta))


def ry_gate(theta=np.pi, device='cpu', dtype=tc.complex64):
    # exp(-i*theta * y/2)
    gate = tc.eye(2)*np.cos(theta/2) - 1j * np.sin(theta/2)*pauli_y
    return Gate(gate.to(device).to(dtype), label=('RY', theta))


def rz_gate(theta=np.pi, device='cpu', dtype=tc.complex64):
    # exp(-i*theta * z/2)
    gate = tc.eye(2)*np.cos(theta/2) - 1j * np.sin(theta/2)*pauli_z
    return Gate(gate.to(device).to(dtype), label=('RZ', theta))


def cz_gate(device='cpu', dtype=tc.complex64):
    gate = tc.eye(4)
    gate[-1, -1] = -1
    return Gate(gate.to(device).to(dtype), label='CZ')


# practical gates with label

def hadamard(device='cpu', dtype=tc.complex64):
    gate = tc.tensor([[1, 1], [1, -1]], device=device, dtype=dtype) / np.sqrt(2)
    return Gate(gate, label='H')


def phase_shift(theta, device='cpu', dtype=tc.complex64):
    gate = tc.eye(2, device=device, dtype=dtype)
    gate[1, 1] = tc.exp(tc.tensor(theta * 1j))
    return Gate(gate)


def not_gate(device='cpu', dtype=tc.complex64):
    # this gate will be discarded
    gate = tc.tensor([[0, 1], [1, 0]], device=device, dtype=dtype)
    return Gate(gate, label='NOT')


def x_gate(device='cpu', dtype=tc.complex64):
    return Gate(pauli_x.to(device).to(dtype), label='X')


def y_gate(device='cpu', dtype=tc.complex64):
    return Gate(pauli_y.to(device).to(dtype), label='Y')


def z_gate(device='cpu', dtype=tc.complex64):
    return Gate(pauli_z.to(device).to(dtype), label='Z')


def swap_gate(device='cpu', dtype=tc.complex64):
    gate = tc.zeros(4, 4, device=device, dtype=dtype)
    gate[0, 0] = 1
    gate[1, 2] = 1
    gate[2, 1] = 1
    gate[3, 3] = 1
    return Gate(gate, label='SWAP')

# gate without label


def rand_gate(dim=2, device='cpu', dtype=tc.complex64, requires_grad=False):
    # Haar random gate
    tmp_tensor = tc.randn(dim, dim, device=device, dtype=dtype)
    q, r = tc.linalg.qr(tmp_tensor)
    sign_matrix = tc.sign(tc.real(tc.diag(r)))
    gate = tc.einsum('ij,j->ij', q, sign_matrix)
    gate = Gate(gate)
    gate.requires_grad = requires_grad
    return gate


def noise_gate_single_gaussian(strength=0.01, device='cpu', dtype=tc.complex64):
    theta = tc.randn(1, device=device) * (strength ** 0.5)
    direction = tc.randn(3, device=device)
    direction = direction / direction.norm()
    sigma = direction[0] * pauli_x + direction[1] * pauli_y + direction[2] * pauli_z
    gate_matrix = tc.cos(theta)*tc.eye(2) + tc.sin(theta) * 1j * sigma
    gate_matrix = gate_matrix.to(device).to(dtype)
    return Gate(gate_matrix)



def time_evolution(hamiltonian, time, device='cpu', dtype=tc.complex64):
    hamiltonian = hamiltonian.to(device).to(dtype)
    if hamiltonian.is_sparse:
        gate = tc.matrix_exp(-1j * hamiltonian.to_dense() * time)
        gate = gate.to_sparse()
    else:
        gate = tc.matrix_exp(-1j * hamiltonian * time)
    return Gate(gate)



# Copyright (C) <2022>  <Zheng-Zhi Sun>
# This file is part of QuantumIntelligence.
# QuantumIntelligence is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Foobar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Foobar. If not, see <https://www.gnu.org/licenses/>.

import random
from collections import Counter
import numpy as np
import torch as tc
from QuantumIntelligence.QuantumSimulator.Circuit import Circuit


# [1, 0] is |0>, [0, 1] is |1>
# do not use fake functions, it will be removed soon
# please apply gate on position as list(range(?)) as much as possible, this will make it faster
# please control gate on position as list(range(?, n_qubit)) as much as possible, this will make it faster

class SimulatorProcess:
    def __init__(self, n_qubit, device='cuda:0', dtype=tc.complex64, rand_seed=1, fast_mode=True, state_acc=1e-3):
        self.n_qubit = n_qubit
        self.device = device
        self.dtype = dtype
        self.rand_seed = rand_seed
        self._state = None
        self.initialize_state()
        self.tmp_gate = None
        self.shape = (2,) * self.n_qubit
        self.chi = 2 ** self.n_qubit
        self.hamiltonian = None
        self.inverse = False
        self.fast_mode = fast_mode
        self.state_acc = state_acc
        self.circuit = Circuit(n_qubit, device=device, dtype=dtype)

    def initialize_state(self):
        self._state = tc.zeros(2 ** self.n_qubit, device=self.device, dtype=self.dtype)
        self._state[0] = 1
        self._state = self._state.view((2,) * self.n_qubit)

    def simulate(self, clear_circuit=True, fast_mode=None, state_acc=None):
        self.circuit.regularize_all_position()

        if fast_mode is None:
            fast_mode = self.fast_mode

        if state_acc is None:
            state_acc = self.state_acc

        if len(self.circuit) > 0:

            # check auto grad
            if fast_mode:
                flag = self.circuit.requires_grad
                for ff in flag:
                    if ff:
                        raise ValueError('fast mode does not support auto grad')

            # check state
            if tc.abs(self._state.norm() - 1) > state_acc:
                raise ValueError('the norm of state is not 1. The diff is ', np.abs(self._state.cpu().norm() - 1))
            self._state = self._state.view(self.shape)
            for ii in range(len(self.circuit)):
                cc = self.circuit[ii]
                if fast_mode:
                    if cc[0].is_fast:
                        self.act_single_gate_fast(label=cc[0].label, position=cc[1], control=cc[2])
                    else:
                        self.act_single_gate(gate=cc[0], position=cc[1], control=cc[2], fast_mode=fast_mode)
                else:
                    self.act_single_gate(gate=cc[0], position=cc[1], control=cc[2], fast_mode=fast_mode)
        if clear_circuit:
            self.circuit = Circuit(self.n_qubit, self.device, self.dtype)

    @property
    def state(self):
        return self._state.clone()

    @state.setter
    def state(self, state):
        self._state = state.clone().reshape(self.shape).to(self.device).to(self.dtype)

    def act_single_gate_fast(self, label, position, control=None):
        if label == 'asdasd':
            pass
        else:
            raise ValueError('The label ', label, ' has not been supported')

    def act_single_gate(self, gate, position, control=None, fast_mode=False):
        # the position and control need to be canonized

        if control is None:
            control = []
        # gate can be sparse, but there seems to be no speedup
        # one should be careful when add inverse gates
        m_p = len(position)
        m_c = len(control)
        old_position = position + control
        new_position = list(range(m_p)) + list(range(-m_c, 0))
        if gate.inverse:
            tmp_gate = gate.tensor.conj().t()
        else:
            tmp_gate = gate.tensor
        self._state = self._state.movedim(old_position, new_position).contiguous().view(2 ** m_p, -1, 2 ** m_c)

        if fast_mode:
            # auto grad will fail
            self._state[:, :, -1] = tmp_gate.mm(self._state[:, :, -1])
        else:

            # The reason to introduce tmp_state is for the auto grad

            tmp_state = self._state.new_empty(self._state.size())
            tmp_state[:, :, -1] = tmp_gate.mm(self._state[:, :, -1])
            tmp_state[:, :, :-1] = self._state[:, :, :-1]
            self._state = tmp_state

        self._state = self._state.view(self.shape).movedim(new_position, old_position)

    def sampling(self, n_shots=1024, position=None, basis=None, if_print=True, rand_seed=None):
        if rand_seed is None:
            rand_seed = self.rand_seed
        if rand_seed is not None:
            random.seed(rand_seed)
        if basis is not None:
            tmp_state = self.change_measure_basis(position, basis)
        else:
            tmp_state = self._state.clone().detach()
        if position is None:
            position = list(range(self.n_qubit))
            weight = tc.abs(tmp_state.contiguous().view(-1)) ** 2
            m_p = len(position)
        else:
            m_p = len(position)
            state_tmp = tmp_state.movedim(position, list(range(m_p))).contiguous().view(2 ** m_p, -1)
            weight = tc.abs(tc.einsum('ab,ba->a', state_tmp, state_tmp.conj().t()))

        population = list()
        for pp in range(2 ** m_p):
            element = bin(pp)[2:]
            element = (m_p - len(element)) * '0' + element
            population.append(element)

        res = Counter(random.choices(population, weight, k=n_shots))
        if if_print:
            for key in res.keys():
                print(key, res[key])
        return res

    @staticmethod
    def count_sample(res, ss, position, if_print=True):
        new_res = dict()
        for key in res.keys():
            flag = True
            for pp in position:
                if key[pp] != ss[position.index(pp)]:
                    flag = False
            if flag:
                new_res[key] = res[key]
        if if_print:
            for key in new_res.keys():
                print(key, new_res[key])
        return new_res

    def change_measure_basis(self, position, basis):
        if position is None:
            position = list(range(self.n_qubit))
        x_basis = tc.tensor([[1, 1], [1, -1]], device=self.device, dtype=self.dtype) / np.sqrt(2)
        y_basis = tc.tensor([[1, 1], [-1j, 1j]], device=self.device, dtype=self.dtype) / np.sqrt(2)
        tmp_state = self._state.clone().detach()
        for nn in range(len(position)):
            pp = position[nn]
            if basis[nn] == 'x':
                tmp_state = tc.einsum('abc,bd->adc', tmp_state.reshape(2 ** pp, -1, 2 ** (self.n_qubit - pp - 1)), x_basis)
            elif basis[nn] == 'y':
                tmp_state = tc.einsum('abc,bd->adc', tmp_state.reshape(2 ** pp, -1, 2 ** (self.n_qubit - pp - 1)), y_basis)
            tmp_state = tmp_state.view(self.shape).contiguous()
        return tmp_state

    def collapse(self, position, basis=None):
        m_p = len(position)
        if basis is None:
            basis = '0' * m_p
        if m_p != len(basis):
            raise ValueError('error in extend, check position')
        index = int(basis, 2)
        new_position = list(range(-m_p, 0))
        tmp_state = self.state.movedim(position, new_position).reshape(-1, 2 ** m_p)
        return tmp_state[:, index]

    def fake_local_measure(self, position, operator):

        assert position == sorted(position), 'warning!!! The position should be in descending order.'
        index_contract = list(range(self.n_qubit))
        for pp in position:
            index_contract.remove(pp)
        reduce_rho = tc.tensordot(self._state, self._state.conj(), dims=[index_contract, index_contract])
        measure_result = tc.einsum('ab,ba->', reduce_rho, operator)
        # print(reduce_rho)
        return measure_result

    def fake_local_rho(self, position):

        assert position == sorted(position), 'warning!!! The position should be in descending order.'
        index_contract = list(range(self.n_qubit))
        for pp in position:
            index_contract.remove(pp)
        reduce_rho = tc.tensordot(self._state, self._state.conj(), dims=[index_contract, index_contract])
        return reduce_rho.reshape(2**len(position), -1)

    def fake_measure_circuit(self, circuit):
        if len(self.circuit) > 0:
            print('warning, there are gates that are not simulated before this fake measure. Measured results are based on the current state')
        backup_circuit = self.circuit
        backup_state = self.state
        self.circuit = circuit
        self.simulate()
        fidelity = tc.einsum('a,a->', backup_state.reshape(-1), self._state.reshape(-1).conj())
        self.circuit = backup_circuit
        self._state = backup_state
        return fidelity





# Copyright (C) <2022>  <Zheng-Zhi Sun>
# This file is part of QuantumIntelligence.
# QuantumIntelligence is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Foobar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Foobar. If not, see <https://www.gnu.org/licenses/>.

import torch as tc
import numpy as np
from QuantumIntelligence.TEBD import TNclass
from QuantumIntelligence.BasicFunSZZ import wheel_function as wf


class MPS(TNclass.TensorNetwork):
    def __init__(self, n_site=4, bond_limit=32, dv=2,
                 dtype=tc.complex128, device='cpu',
                 cut_off=1e-14):
        # Prepare parameters
        TNclass.TensorNetwork.__init__(self)
        self.n_site = n_site
        self.bond_limit = bond_limit
        self.dv = dv
        self.cut_off = cut_off
        self.n_site = n_site
        self.regular_center = None
        self.device = device
        self.dtype = dtype

    def initialize_mps(self, method=None, rand_seed=None):
        if method is None:
            method = 'rand'
        if rand_seed is not None:
            tc.manual_seed(rand_seed)
        self.tensor_data = list()

        dv = self.dv

        for nn in range(self.n_site):
            if method == 'rand':
                self.tensor_data.append(tc.rand((1, dv, 1), device=self.device, dtype=self.dtype))
            elif method == 'ones':
                self.tensor_data.append(tc.ones((1, dv, 1), device=self.device, dtype=self.dtype))


    def mps_regularization(self, regular_center, if_force=True):
        tmp_bond_limit = self.bond_limit
        regular_center = regular_center % self.n_site
        if self.regular_center is None:
            self.regular_center = 0
            if if_force:
                self.bond_limit = np.inf
            while self.regular_center < self.n_site-1:
                self.move_regular_center2next()
            self.bond_limit = tmp_bond_limit
        while self.regular_center < regular_center:
            self.move_regular_center2next()
        while self.regular_center > regular_center:
            self.move_regular_center2forward()

    def move_regular_center2next(self, norm_mode=True, cut_off=None, bond_limit=None):
        if cut_off is None:
            cut_off = self.cut_off
        if bond_limit is None:
            bond_limit = self.bond_limit

        tensor_index = self.regular_center
        tmp_tensor = self.tensor_data[tensor_index]
        d0, dv, d1 = tmp_tensor.shape
        u, s, v = wf.safe_svd(tmp_tensor.reshape(d0*dv, d1))
        if norm_mode:
            s = s/s.norm()
        s = s[s > cut_off]
        dm = min(s.numel(), bond_limit)
        u = u[:, :dm]
        s = s[:dm]
        v = v[:dm, :]

        self.tensor_data[tensor_index] = u.reshape(d0, dv, -1)
        self.tensor_data[tensor_index + 1] = tc.einsum('b,bc,cde->bde', s, v, self.tensor_data[tensor_index + 1])
        self.regular_center += 1

    def move_regular_center2forward(self, norm_mode=True, cut_off=None, bond_limit=None):
        if cut_off is None:
            cut_off = self.cut_off
        if bond_limit is None:
            bond_limit = self.bond_limit

        tensor_index = self.regular_center
        tmp_tensor = self.tensor_data[tensor_index]
        d0, dv, d1 = tmp_tensor.shape
        u, s, v = wf.safe_svd(tmp_tensor.reshape(d0, dv * d1))
        if norm_mode:
            s = s/s.norm()
        s = s[s > cut_off]
        dm = min(s.numel(), bond_limit)
        u = u[:, :dm]
        s = s[:dm]
        v = v[:dm, :]
        self.tensor_data[tensor_index] = v.reshape(dm, dv, d1)
        self.tensor_data[tensor_index - 1] = tc.einsum(
            'abc,cd,d->abd', self.tensor_data[tensor_index - 1], u, s).reshape(-1, dv, dm)
        self.regular_center -= 1




# Copyright (C) <2022>  <Zheng-Zhi Sun>
# This file is part of QuantumIntelligence.
# QuantumIntelligence is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Foobar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Foobar. If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import torch as tc
import copy
from QuantumIntelligence.QuantumSimulator import Gate
from QuantumIntelligence.BasicFunSZZ import tensor_trick as tt


# [1, 0] is |0>, [0, 1] is |1>
# do not use fake functions, it will be removed soon
# please apply gate on position as list(range(?)) as much as possible, this will make it faster
# please control gate on position as list(range(?, n_qubit)) as much as possible, this will make it faster

class Circuit:

    def __init__(self, n_qubit, device='cpu', dtype=tc.complex64, size=None):
        self.n_qubit = n_qubit
        self.device = device
        self.dtype = dtype
        self.size = size
        self.gate_list = list()
        self.position_list = list()
        self.control_list = list()
        self.barrier_index = list()
        self.noised = False
        self.ready4experiment = False

    @property
    def requires_grad(self):
        flag = list()
        for gg in self.gate_list:
            flag.append(gg.requires_grad)
        return flag

    @requires_grad.setter
    def requires_grad(self, requires_grad):
        if isinstance(requires_grad, bool):
            for gg in self:
                gg.requires_grad = requires_grad
        else:
            for nn in range(len(self)):
                self.gate_list[nn].requires_grad = requires_grad[nn]

    # core functions start

    def compose(self, circuit, position=None, control=None, inverse=False):
        if isinstance(circuit, Circuit):
            tmp_circuit = copy.deepcopy(circuit)
            if inverse:
                tmp_circuit.inv()
            self.__extend(tmp_circuit, position=position, control=control)
        elif isinstance(circuit, Gate.Gate):
            if position is None:
                raise ValueError('position is needed when compose gate')
            self.add_single_gate(gate=circuit, position=position, control=control, inverse=inverse)
        else:
            raise ValueError('input is not Gate or Circuit')

    def __extend(self, circuit, position=None, control=None):
        if not isinstance(circuit, Circuit):
            raise TypeError('this function is used to extend circuit. Use append or compose to add single gate')
        if control is None:
            control = []
        if position is None:
            if self.n_qubit == circuit.n_qubit:
                for ii in range(len(circuit)):
                    cc = circuit[ii]
                    self.append(gate=cc[0], position=cc[1], control=cc[2])
            else:
                raise ValueError('error in extend, position is needed')
        elif len(position) == circuit.n_qubit:
            for ii in range(len(circuit)):
                cc = circuit[ii]
                new_p = list()
                new_c = list()
                for oo in cc[1]:
                    new_p.append(position[oo])
                for oo in cc[2]:
                    new_c.append(position[oo])
                for oo in control:
                    new_c.append(oo)
                self.append(gate=cc[0], position=new_p.copy(), control=new_c.copy())
        else:
            raise ValueError('error in extend, check position')

    def add_single_gate(self, gate, position, control=None, inverse=False):
        # gate can be sparse, but there seems to be no speedup
        # one should be careful when add inverse gates
        # the gate will be cloned and detached by default

        # do not change the following code, or some bugs would occur in circuit.ch
        gate = Gate.Gate(gate, independent=True)
        # gate = Gate.Gate(gate, independent=True).to(self.device).to(self.dtype)

        if tt.have_same_iterable(position, control):
            raise ValueError('position and control have same qubits')
        self.position_list.append(position)
        if control is None:
            control = []
        self.control_list.append(control)
        if inverse:
            gate.inv()
        self.gate_list.append(gate)

    def append(self, gate, position, control=None, inverse=False):
        # this is just another name of add single gate
        self.add_single_gate(gate, position=position, control=control, inverse=inverse)

    def repeat_append(self, gate, position_list, inverse=False):
        for position in position_list:
            self.add_single_gate(gate, position=position, inverse=inverse)

    def inv(self):
        self.gate_list.reverse()
        self.position_list.reverse()
        self.control_list.reverse()
        for gate in self.gate_list:
            gate.inv()

    def return_inv(self):
        tmp_circuit = copy.deepcopy(self)
        tmp_circuit.gate_list.reverse()
        tmp_circuit.position_list.reverse()
        tmp_circuit.control_list.reverse()
        for gate in tmp_circuit.gate_list:
            gate.inv()
        return tmp_circuit

    def to(self, device_or_dtype):
        for cc in self.gate_list:
            cc.to(device_or_dtype)
        return self

    def square(self):
        for gate in self.gate_list:
            gate.square()

    def regularize_position(self, index):

        self.position_list[index] = list(map(lambda x: x % self.n_qubit, self.position_list[index]))
        self.control_list[index] = list(map(lambda x: x % self.n_qubit, self.control_list[index]))

    def regularize_all_position(self):
        for ii in range(len(self)):
            self.regularize_position(ii)

    def __getitem__(self, index):
        return self.gate_list[index], self.position_list[index], self.control_list[index]

    def __len__(self):
        return len(self.gate_list)

    def __add__(self, circuit):

        if circuit.n_qubit != self.n_qubit:
            raise ValueError('error in +, n_qubit not equal')

        tmp_circuit = Circuit(self.n_qubit, device=self.device, dtype=self.dtype)
        tmp_circuit.gate_list = self.gate_list + circuit.gate_list
        tmp_circuit.position_list = self.position_list + circuit.position_list
        tmp_circuit.control_list = self.control_list + circuit.control_list
        return tmp_circuit

    def __copy__(self):
        raise NameError('I do not permit copy for now. Please use deepcopy().')

    def pop(self, index):
        self.gate_list.pop(index)
        self.position_list.pop(index)
        self.control_list.pop(index)

    def add_noise_gate(self, strength=0.01):
        for ii in range(len(self)):
            if self.control_list[ii] is not None:
                tmp_noise = Gate.noise_gate_single_gaussian(strength=strength, device=self.device, dtype=self.dtype)
                self.gate_list[ii].tensor = self.gate_list[ii].tensor.mm(tmp_noise.tensor)
        self.noised = True

    def cal_barrier(self):
        self.regularize_all_position()
        position_list = list(range(self.n_qubit))
        flag = False
        for ii in range(len(self)):
            for nn in self.position_list[ii]:
                if nn not in position_list:
                    flag = True
            for nn in self.control_list[ii]:
                if nn not in position_list:
                    flag = True

            if flag:
                self.barrier_index.append(ii)
                position_list = list(range(self.n_qubit))
                flag = False
            for nn in self.position_list[ii]:
                position_list.remove(nn)
            for nn in self.control_list[ii]:
                position_list.remove(nn)

    def regularize_gates(self):
        new_gate_list = []
        new_position_list = []
        new_control_list = []
        while len(self) > 0:
            index_list_strict = list(range(self.n_qubit))
            index_list_loose = list(range(self.n_qubit))
            ii = 0
            while ii < len(self):

                flag = True
                for nn in self.position_list[ii]:
                    if nn not in index_list_strict:
                        flag = False
                for nn in self.control_list[ii]:
                    if nn not in index_list_loose:
                        flag = False

                if flag:
                    for nn in self.position_list[ii]:
                        if nn in index_list_strict:
                            index_list_strict.remove(nn)
                        if nn in index_list_loose:
                            index_list_loose.remove(nn)
                    for nn in self.control_list[ii]:
                        if nn in index_list_strict:
                            index_list_strict.remove(nn)
                        if nn in index_list_loose:
                            index_list_loose.remove(nn)
                    new_gate_list.append(self.gate_list[ii])
                    new_position_list.append(self.position_list[ii])
                    new_control_list.append(self.control_list[ii])
                    self.pop(ii)
                else:
                    for nn in self.position_list[ii]:
                        if nn in index_list_strict:
                            index_list_strict.remove(nn)
                        if nn in index_list_loose:
                            index_list_loose.remove(nn)
                    for nn in self.control_list[ii]:
                        if nn in index_list_strict:
                            index_list_strict.remove(nn)
                    ii = ii + 1
        self.gate_list = new_gate_list
        self.position_list = new_position_list
        self.control_list = new_control_list

    # core functions end
    # practical functions start

    # universal gate set from experiment

    def rx_gate(self, position, theta=np.pi, control=None):
        tmp_gate = Gate.rx_gate(theta=theta, device=self.device, dtype=self.dtype)
        for pp in position:
            self.add_single_gate(tmp_gate, position=[pp], control=control)

    def ry_gate(self, position, theta=np.pi, control=None):
        tmp_gate = Gate.ry_gate(theta=theta, device=self.device, dtype=self.dtype)
        for pp in position:
            self.add_single_gate(tmp_gate, position=[pp], control=control)

    def rz_gate(self, position, theta=np.pi, control=None):
        tmp_gate = Gate.rz_gate(theta=theta, device=self.device, dtype=self.dtype)
        for pp in position:
            self.add_single_gate(tmp_gate, position=[pp], control=control)

    def cz_gate(self, position):
        tmp_gate = Gate.cz_gate(device=self.device, dtype=self.dtype)
        for pp in position:
            self.add_single_gate(tmp_gate, position=pp)

    # decompose circuit to experimental gates

    def decompose2experimental_gate(self):
        # global phases are ignored

        old_gate_list = self.gate_list
        old_position_list = self.position_list
        old_control_list = self.control_list

        self.gate_list = list()
        self.position_list = list()
        self.control_list = list()

        experimental_gate_set = ('RX', 'RY', 'RZ', 'CZ')
        supported_label = ('X', 'Y', 'Z')
        for ii in range(len(old_gate_list)):
            tmp_label = old_gate_list[ii].label
            if isinstance(tmp_label, tuple):
                tmp_label = tmp_label[0]
            tmp_position = old_position_list[ii]
            tmp_control = old_control_list[ii]

            if tmp_label is None:
                ValueError('The ', str(ii), '-th gate has no label')
            if len(tmp_control) > 1:
                ValueError('Do not support multi controlled ', tmp_label)

            elif tmp_label in experimental_gate_set:
                if len(tmp_control) == 0:
                    self.append(old_gate_list[ii], position=tmp_position)
                elif len(tmp_control) == 1:
                    ValueError('Waiting to support controlled ', tmp_label)
            elif tmp_label in supported_label:
                if tmp_label == 'Z':
                    if len(tmp_control) == 0:
                        self.rz_gate(position=tmp_position, theta=np.pi)
                    elif len(tmp_control) == 1:
                        self.cz_gate(position=[tmp_position+tmp_control])
                    else:
                        ValueError('tired')
                elif tmp_label == 'X':

                    if len(tmp_control) == 0:
                        self.rx_gate(position=tmp_position, theta=np.pi)
                    elif len(tmp_control) == 1:
                        self.rz_gate(position=tmp_position, theta=np.pi)
                        self.ry_gate(position=tmp_position, theta=np.pi/2)
                        self.cz_gate(position=[tmp_position+tmp_control])
                        self.ry_gate(position=tmp_position, theta=-np.pi / 2)
                        self.rz_gate(position=tmp_position, theta=-np.pi)

                    else:
                        ValueError('tired')
                elif tmp_label == 'Y':
                    if len(tmp_control) == 0:
                        self.ry_gate(position=tmp_position, theta=np.pi)
                    elif len(tmp_control) == 1:
                        self.rz_gate(position=tmp_position, theta=3*np.pi/2)
                        self.ry_gate(position=tmp_position, theta=-np.pi/2)
                        self.cz_gate(position=[tmp_position+tmp_control])
                        self.ry_gate(position=tmp_position, theta=np.pi/2)
                        self.rz_gate(position=tmp_position, theta=-3*np.pi/2)
                    else:
                        ValueError('tired')

    # useful gates and circuits

    def x_gate(self, position, control=None):
        tmp_gate = Gate.x_gate(device=self.device, dtype=self.dtype)
        for pp in position:
            self.add_single_gate(tmp_gate, position=[pp], control=control)

    def y_gate(self, position, control=None):
        tmp_gate = Gate.y_gate(device=self.device, dtype=self.dtype)
        for pp in position:
            self.add_single_gate(tmp_gate, position=[pp], control=control)

    def z_gate(self, position, control=None):
        tmp_gate = Gate.z_gate(device=self.device, dtype=self.dtype)
        for pp in position:
            self.add_single_gate(tmp_gate, position=[pp], control=control)

    def hadamard(self, position, control=None):
        tmp_gate = Gate.hadamard(device=self.device, dtype=self.dtype)
        for pp in position:
            self.add_single_gate(tmp_gate, position=[pp], control=control)

    def phase_shift(self, position, theta=tc.pi, control=None):
        for pp in position:
            tmp_gate = Gate.phase_shift(theta, device=self.device, dtype=self.dtype)
            self.add_single_gate(tmp_gate, position=[pp], control=control)

    def not_gate(self, position, control=None):
        for pp in position:
            tmp_gate = Gate.not_gate(device=self.device, dtype=self.dtype)
            self.add_single_gate(tmp_gate, position=[pp], control=control)

    def rand_gate(self, dim, position, control=None, requires_grad=False):
        tmp_gate = Gate.rand_gate(dim, device=self.device, dtype=self.dtype, requires_grad=requires_grad)
        self.add_single_gate(tmp_gate, position=position, control=control)

    def swap_gate(self, position, control=None):
        if len(position) != 2:
            print('wrong use')
        else:
            tmp_gate = Gate.swap_gate(device=self.device, dtype=self.dtype)
            self.add_single_gate(tmp_gate, position=position, control=control)

    def time_evolution(self, hamiltonian, time, position, control=None):
        tmp_gate = Gate.time_evolution(hamiltonian, time, device=self.device, dtype=self.dtype)
        self.add_single_gate(tmp_gate, position=position, control=control)

    def qft(self, position, control=None, inverse=False):
        if control is None:
            control = []
        tmp_circuit = qft(self.n_qubit, position=position, control=control, inverse=inverse, device=self.device, dtype=self.dtype)
        self.__extend(tmp_circuit)

    def ch(self, unitary, position_phi, position_c, control=None, inverse=False):
        tmp_circuit = ch(self.n_qubit, unitary, position_phi, position_c, control, inverse, device=self.device, dtype=self.dtype)
        self.__extend(tmp_circuit)

    def qpe(self, unitary, position_phi, position_qpe, control=None, inverse=False):
        tmp_circuit = qpe(self.n_qubit, unitary, position_phi, position_qpe, control, inverse, device=self.device, dtype=self.dtype)
        self.__extend(tmp_circuit)

    def add_one(self, position=None, control=None, inverse=False):
        tmp_circuit = add_one(self.n_qubit, position, control, inverse, device=self.device, dtype=self.dtype)
        self.__extend(tmp_circuit)

    def qhc(self, unitary, position_phi, position_qpe, position_f, n_f=None,
            control=None, inverse=False):
        tmp_circuit = qhc(self.n_qubit, unitary, position_phi, position_qpe, position_f,
                          n_f, control, inverse, device=self.device, dtype=self.dtype)
        self.__extend(tmp_circuit)

    def quantum_coin(self, unitary, position_phi, position_coin, control=None, inverse=False):
        tmp_circuit = quantum_coin(self.n_qubit, unitary, position_phi, position_coin,
                                   control, inverse, device=self.device, dtype=self.dtype)
        self.__extend(tmp_circuit)

    def qdc(self, unitary, position_phi, position_coin, position_f,
            n_f=None, control=None, inverse=False):
        tmp_circuit = qdc(self.n_qubit, unitary, position_phi, position_coin, position_f,
                          n_f, control, inverse, device=self.device, dtype=self.dtype)
        self.__extend(tmp_circuit)


def qft(n_qubit, position, control=None, inverse=False, device='cpu', dtype=tc.complex64):
    if control is None:
        control = []
    tmp_circuit = Circuit(n_qubit, device, dtype)
    m_qft = len(position)
    perm = list(range(m_qft))
    perm.reverse()
    theta_list = []
    theta = 2 * np.pi
    if inverse:
        theta = -theta
    for mm in range(m_qft + 1):
        theta_list.append(theta)
        theta = theta / 2
    for mm in range(m_qft):
        tmp_circuit.hadamard([position[mm]])
        for nn in range(mm + 1, m_qft):
            tmp_circuit.phase_shift(theta_list[nn - mm + 1], [position[mm]], control=[position[nn]] + control)
    for mm in range(m_qft // 2):
        tmp_circuit.swap_gate([position[mm], position[- mm - 1]], control=control)
    if inverse:
        tmp_circuit.inv()
    return tmp_circuit


def ch(n_qubit, unitary, position_phi, position_c, control=None, inverse=False, device='cpu', dtype=tc.complex64):
    if control is None:
        control = []
    tmp_circuit = Circuit(n_qubit, device, dtype)
    if isinstance(unitary, Gate.Gate):
        tmp_gate = Gate.Gate(unitary)
    elif isinstance(unitary, Circuit):
        tmp_gate = copy.deepcopy(unitary)
    else:
        raise TypeError('the unitary should be a gate')
    m_fch = len(position_c)
    for mm in range(m_fch):
        tmp_circuit.compose(tmp_gate, position_phi, [position_c[- mm - 1]] + control, inverse)
        tmp_gate.square()
    if inverse:
        tmp_circuit.inv()
    return tmp_circuit


def qpe(n_qubit, unitary, position_phi, position_qpe, control=None, inverse=False, device='cpu', dtype=tc.complex64):
    if control is None:
        control = []
    tmp_circuit = Circuit(n_qubit, device, dtype)
    tmp_circuit.hadamard(position_qpe, control=control)
    tmp_circuit.ch(unitary, position_phi, position_qpe, control, False)
    tmp_circuit.qft(position_qpe, control=control, inverse=True)
    if inverse:
        tmp_circuit.inv()
    return tmp_circuit


def qhc(n_qubit, unitary, position_phi, position_qpe, position_f, n_f=None,
        control=None, inverse=False, device='cpu', dtype=tc.complex64):
    tmp_circuit = Circuit(n_qubit, device, dtype)
    if control is None:
        control = []
    if n_f is None:
        n_f = 2 ** (len(position_f) - 1) - 1
    if n_f > (2 ** len(position_f) - 1) / 2:
        print('warning, n_f is too large')
    for nn in range(n_f):
        tmp_circuit.not_gate(position_f, control)
        tmp_circuit.qpe(unitary, position_phi, position_qpe,
                        control=position_f + control, inverse=False)
        # sample(1000, position_qpe)
        tmp_circuit.not_gate(position_f, control)
        tmp_circuit.add_one(position_f, [position_qpe[0]] + control, inverse=False)
        tmp_circuit.not_gate(position_f, control)
        tmp_circuit.qpe(unitary, position_phi, position_qpe,
                        control=position_f + control, inverse=True)
        tmp_circuit.not_gate(position_f, control)
        tmp_circuit.add_one(position_f, control=control, inverse=False)
        tmp_circuit.not_gate(position_qpe, control)
        tmp_circuit.add_one(position_f, control=position_qpe + control, inverse=True)
        tmp_circuit.not_gate(position_qpe, control)
    if inverse:
        tmp_circuit.inv()
    return tmp_circuit


def add_one(n_qubit, position=None, control=None, inverse=False, device='cpu', dtype=tc.complex64):
    tmp_circuit = Circuit(n_qubit, device, dtype)
    if control is None:
        control = []
    if position is None:
        position = list(range(n_qubit))
    m_a = len(position)
    for mm in range(m_a):
        tmp_circuit.not_gate([position[mm]], control=position[mm + 1:] + control)
    if inverse:
        tmp_circuit.inv()
    return tmp_circuit


def quantum_coin(n_qubit, unitary, position_phi, position_coin, control=None, inverse=False,
                 device='cpu', dtype=tc.complex64):
    tmp_circuit = Circuit(n_qubit, device, dtype)
    if control is None:
        control = []
    tmp_circuit.hadamard(position_coin, control)
    tmp_circuit.compose(unitary, position_phi, position_coin)
    tmp_circuit.not_gate(position_coin, control)
    tmp_circuit.compose(unitary, position_phi, position_coin, True)
    # self.phase_shift(-eig_value, position_coin, control)
    tmp_circuit.not_gate(position_coin, control)
    tmp_circuit.hadamard(position_coin, control)
    if inverse:
        tmp_circuit.inv()
    return tmp_circuit


def qdc(n_qubit, unitary, position_phi, position_coin, position_f, n_f=None,
        control=None, inverse=False, device='cpu', dtype=tc.complex64):
    tmp_circuit = Circuit(n_qubit, device, dtype)
    if control is None:
        control = []
    if n_f is None:
        n_f = 2 ** (len(position_f) - 1) - 1
    if n_f > (2 ** len(position_f) - 1):
        print('warning, n_f is too large')
    for nn in range(n_f):
        tmp_circuit.not_gate(position_f, control)
        tmp_circuit.quantum_coin(unitary, position_phi, position_coin, control=position_f + control, inverse=False)
        tmp_circuit.not_gate(position_f, control)
        tmp_circuit.add_one(position_f, position_coin + control, inverse=False)
    if inverse:
        tmp_circuit.inv()
    return tmp_circuit



# Copyright (C) <2022>  <Zheng-Zhi Sun>
# This file is part of QuantumIntelligence.
# QuantumIntelligence is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Foobar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Foobar. If not, see <https://www.gnu.org/licenses/>.

import scipy.io as io
import torch as tc
import numpy as np
import math


def safe_svd(matrix):
    if tc.any(tc.isnan(matrix)):
        print('!!!!!!!!!!')
        print(matrix)
        raise ValueError('there is nan in matrix')
    try:
        u, s, v = tc.linalg.svd(matrix, full_matrices=False)
    except RuntimeError:
        device = matrix.device
        dtype = matrix.dtype
        u, s, v = np.linalg.svd(matrix.cpu())
        u = tc.from_numpy(u).to(device).to(dtype)
        s = tc.from_numpy(s).to(device).to(dtype)
        v = tc.from_numpy(v).to(device).to(dtype)
    if tc.any(tc.isnan(u)) or tc.any(tc.isnan(v)) or tc.any(tc.isnan(s)):
        device = matrix.device
        dtype = matrix.dtype
        u, s, v = np.linalg.svd(matrix.cpu())
        u = tc.from_numpy(u).to(device).to(dtype)
        s = tc.from_numpy(s).to(device).to(dtype)
        v = tc.from_numpy(v).to(device).to(dtype)

    return u, s.real, v


def load_mat(file_name, file_path='./'):
    merged_name = file_path + file_name
    data_loaded = io.loadmat(merged_name)
    return data_loaded


def np_kron(*args):
    tmp = args[0]
    for nn in range(1, len(args)):
        tmp = np.kron(tmp, args[nn])
    return tmp


def outer_parallel(a, *matrix):
    # need optimization
    for b in matrix:
        a = (a.repeat(b.shape[1], 1).reshape(a.shape + (-1,))
             * b.repeat(a.shape[1], 0).reshape(a.shape + (-1,))).reshape(a.shape[0], -1)
    return a


def outer(a, *matrix):
    for b in matrix:
        a = np.outer(a, b).flatten()
    return a


def tensor_contract(a, b, index):
    ndim_a = np.array(a.shape)
    ndim_b = np.array(b.shape)
    order_a = np.arange(len(ndim_a))
    order_b = np.arange(len(ndim_b))
    order_a_contract = np.array(order_a[index[0]]).flatten()
    order_b_contract = np.array(order_b[index[1]]).flatten()
    order_a_hold = np.setdiff1d(order_a, order_a_contract)
    order_b_hold = np.setdiff1d(order_b, order_b_contract)
    hold_shape_a = ndim_a[order_a_hold].flatten()
    hold_shape_b = ndim_b[order_b_hold].flatten()
    return np.dot(
        a.transpose(np.concatenate([order_a_hold, order_a_contract])).reshape(hold_shape_a.prod(), -1),
        b.transpose(np.concatenate([order_b_contract, order_b_hold])).reshape(-1, hold_shape_b.prod())) \
        .reshape(np.concatenate([hold_shape_a, hold_shape_b]))


def tensor_svd(tmp_tensor, index_left='none', index_right='none'):
    tmp_shape = np.array(tmp_tensor.shape)
    tmp_index = np.arange(len(tmp_tensor.shape))
    if index_left == 'none':
        index_right = tmp_index[index_right].flatten()
        index_left = np.setdiff1d(tmp_index, index_right)
    elif index_right == 'none':
        index_left = tmp_index[index_left].flatten()
        index_right = np.setdiff1d(tmp_index, index_left)
    index_right = np.array(index_right).flatten()
    index_left = np.array(index_left).flatten()
    tmp_tensor = tmp_tensor.transpose(np.concatenate([index_left, index_right]))
    tmp_tensor = tmp_tensor.reshape(tmp_shape[index_left].prod(), tmp_shape[index_right].prod())
    u, l, v = np.linalg.svd(tmp_tensor, full_matrices=False)
    return u, l, v


def calculate_mse(image_origin, image_noised):
    return ((image_origin - image_noised) ** 2) / np.prod(image_origin.shape)


def calculate_psnr(image_origin, image_noised):
    # not very correct
    image_origin = image_origin.flatten()
    image_noised = image_noised.flatten()
    return 20 * np.log10(
        np.concatenate((image_origin, image_noised)).max()) - 10 * np.log10(
        ((image_origin - image_noised) ** 2).sum() / np.prod(image_origin.shape))


# QHES related


def get_x_qpe(n):
    rx = -(n + 2)/math.log2(1 - (2/math.pi)**2)
    x = int(rx) + 2
    return x


def get_m_qpe(n):
    x = get_x_qpe(n)
    m = 0
    result_p = 0
    while result_p < 0.8:
        m = m + 1
        tmp_x = math.pi/(2 * m)
        result_p = (math.sin(tmp_x)/tmp_x) ** (2*x)
    return m


def np_kron(*args):
    tmp = args[0]
    for nn in range(1, len(args)):
        tmp = np.kron(tmp, args[nn])
    return tmp

# QHES related end


if __name__ == '__main__':
    print('This is a pure function file. What are you doing?')




# Copyright (C) <2022>  <Zheng-Zhi Sun>
# This file is part of QuantumIntelligence.
# QuantumIntelligence is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Foobar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Foobar. If not, see <https://www.gnu.org/licenses/>.

from collections.abc import Iterable
import torch as tc


def inverse_permutation(perm):
    # perm is a torch tensor
    if not isinstance(perm, tc.Tensor):
        perm = tc.tensor(perm)
    inv = tc.empty_like(perm)
    inv[perm] = tc.arange(perm.size(0), device=perm.device)
    return inv.tolist()


def have_same_iterable(a_list, b_list):
    if isinstance(a_list, Iterable) and isinstance(b_list, Iterable):
        xx = [x for x in a_list if x in b_list]
        if len(xx) > 0:
            return True
        else:
            return False
    else:
        return False


def is_str_same(a_str, b_str):
    flag = True
    if len(a_str) != len(b_str):
        return False
    else:
        for nn in range(len(a_str)):
            if a_str[nn] != b_str[nn]:
                flag = False
        return flag



# Copyright (C) <2022>  <Zheng-Zhi Sun>
# This file is part of QuantumIntelligence.
# QuantumIntelligence is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Foobar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Foobar. If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import torch as tc
from QuantumIntelligence.TEBD import MPSclass
from QuantumIntelligence.BasicFunSZZ import BasicClass, wheel_function as wf


class TEBD(MPSclass.MPS):

    def __init__(self, n_site=4, bond_limit=32, dv=2,
                 dtype=tc.complex128, device='cpu',
                 cut_off=1e-14, tau_init=1e-1, tau_acc=1e-5,
                 tensor_acc=1e-6, bin_max=10000, energy_acc=1e-6):

        MPSclass.MPS.__init__(self, n_site=n_site, bond_limit=bond_limit, dv=dv,
                              dtype=dtype, device=device, cut_off=cut_off)

        # initialize parameters
        self.tau_init = tau_init
        self.tau_acc = tau_acc
        self.tau = tau_init
        self.tensor_acc = tensor_acc
        self.bin_max = bin_max
        self.energy_acc = energy_acc

        # initialize mps
        self.initialize_mps(method='ones')
        self.mps_regularization(0)
        self.tensor_data[0] = self.tensor_data[0] / self.tensor_data[0].norm()
        self.s_list = BasicClass.MultiDimensionDict()
        self.old_tensor_data = self.tensor_data.copy()
        self.energy = None

    @staticmethod
    def calculate_exp_ham(ham, tau):
        return tc.matrix_exp(-tau * ham)

    def evolute_one_site(self, index, exp_ham):
        self.tensor_data[index] = tc.einsum('abc,bd->adc', self.tensor_data[index], exp_ham)
        self.regular_center = None

    def evolute_two_site(self, index, exp_ham):
        index0 = index
        index1 = index + 1
        d0, dv, d1 = self.tensor_data[index0].shape
        d2 = self.tensor_data[index1].shape[-1]
        cut_off = self.cut_off
        tmp_tensor = tc.einsum('abc,cde->abde', self.tensor_data[index0], self.tensor_data[index1])
        new_tensor = tc.einsum('abde,bdij->aije', tmp_tensor, exp_ham)
        # error = (tmp_tensor - new_tensor).norm()

        u, s, v = wf.safe_svd(new_tensor.reshape(d0 * dv, dv * d2))

        s = s / s.norm()
        s = s[s > cut_off]
        dm = len(s)
        u = u[:, :dm]
        v = v[:dm, :]
        self.tensor_data[index0] = u.reshape(d0, dv, -1)
        self.tensor_data[index1] = tc.einsum('b,bc->bc', s, v).reshape(dm, dv, d2)
        self.regular_center = None

    def evolute_three_site(self, index, exp_ham):
        index0 = index
        index1 = index + 1
        index2 = index + 2
        d0, dv, d1 = self.tensor_data[index0].shape
        d2 = self.tensor_data[index2].shape[-1]
        cut_off = self.cut_off
        tmp_tensor = tc.einsum('abc,cde,efg->abdfg', self.tensor_data[index0],
                               self.tensor_data[index1], self.tensor_data[index2])
        new_tensor = tc.einsum('abdfg,bdfijk->aijkg', tmp_tensor, exp_ham)
        # error = (tmp_tensor - new_tensor).norm()
        u, s, v = wf.safe_svd(new_tensor.reshape(d0 * dv, dv * dv * d2))
        s = s / s.norm()
        s = s[s > cut_off]
        dm0 = len(s)
        u = u[:, :dm0]
        v = v[:dm0, :]
        self.tensor_data[index0] = u.reshape(d0, dv, dm0)
        new_tensor = tc.einsum('b,bc->bc', s, v).reshape(dm0 * dv, dv * d2)
        # u, s, v = tc.linalg.svd(new_tensor, full_matrices=False)
        u, s, v = wf.safe_svd(new_tensor)
        s = s / s.norm()
        s = s[s > cut_off]
        dm1 = len(s)
        u = u[:, :dm1]
        v = v[:dm1, :]
        self.tensor_data[index1] = u.reshape(dm0, dv, dm1)
        self.tensor_data[index2] = tc.einsum('b, bc->bc', s, v).reshape(dm1, dv, d2)
        self.regular_center = None

    def evolute_all_once(self, exp_ham):
        n_interval = round(np.log2(exp_ham.numel()) / 2)
        exp_ham = exp_ham.reshape(2 * n_interval * (2, ))
        # print(n_interval)
        for nn in range(n_interval):
            tmp_index = nn
            while (tmp_index + n_interval) <= self.n_site:
                if n_interval == 1:
                    self.evolute_one_site(tmp_index, exp_ham)
                elif n_interval == 2:
                    self.evolute_two_site(tmp_index, exp_ham)
                elif n_interval == 3:
                    self.evolute_three_site(tmp_index, exp_ham)
                tmp_index = tmp_index + n_interval
            self.mps_regularization(0)
            self.tensor_data[0] = self.tensor_data[0] / self.tensor_data[0].norm()

    def evolute_all_two_site_second(self, ham, t):
        exp_ham0 = self.calculate_exp_ham(ham, t/2)
        exp_ham1 = self.calculate_exp_ham(ham, t)
        n_interval = 2
        exp_ham0 = exp_ham0.reshape(2 * n_interval * (2, ))
        exp_ham1 = exp_ham1.reshape(2 * n_interval * (2, ))
        tmp_index = 0
        while (tmp_index + n_interval) <= self.n_site:
            self.evolute_two_site(tmp_index, exp_ham0)
            tmp_index = tmp_index + n_interval
        self.mps_regularization(0)
        self.tensor_data[0] = self.tensor_data[0] / self.tensor_data[0].norm()

        tmp_index = 1
        while (tmp_index + n_interval) <= self.n_site:
            self.evolute_two_site(tmp_index, exp_ham1)
            tmp_index = tmp_index + n_interval
        self.mps_regularization(0)
        self.tensor_data[0] = self.tensor_data[0] / self.tensor_data[0].norm()

        tmp_index = 0
        while (tmp_index + n_interval) <= self.n_site:
            self.evolute_two_site(tmp_index, exp_ham0)
            tmp_index = tmp_index + n_interval
        self.mps_regularization(0)
        self.tensor_data[0] = self.tensor_data[0] / self.tensor_data[0].norm()

    def evolute_all_energy(self, ham_list):
        # self.tau = self.tau_init
        n_iter = 0
        bin_num = 0
        tmp_energy = np.inf
        while self.tau > self.tau_acc:
            self.mps_regularization(0)
            exp_ham_list = list()
            for ham in ham_list:
                exp_ham_list.append(self.calculate_exp_ham(ham, self.tau/2))
            for exp_ham in exp_ham_list:
                self.evolute_all_once(exp_ham)
            exp_ham_list.reverse()
            for exp_ham in exp_ham_list:
                self.evolute_all_once(exp_ham)

            self.energy = self.calculate_energy(ham_list)/self.n_site
            tmp_error = tmp_energy - self.energy
            tmp_energy = self.energy
            if tmp_error < (self.cut_off * self.bond_limit):
                print('error is too small, stop this iteration. energy diff = ' + str(tmp_error))
                tmp_error = 0
            tmp_error = tmp_error / self.tau
            n_iter = n_iter + 1
            bin_num = bin_num + 1
            if (tmp_error < self.tensor_acc) or (bin_num >= self.bin_max):
                self.tau = self.tau / 2
                bin_num = 0
                print('n_iter = ' + str(n_iter) + ' tau = ' + str(self.tau) + ' error = ' + str(tmp_error))
        self.mps_regularization(0)
        self.tensor_data[0] = self.tensor_data[0] / self.tensor_data[0].norm()

    def evolute_all_fidelity(self, ham_list):
        # self.tau = self.tau_init
        n_iter = 0
        bin_num = 0
        while self.tau > self.tau_acc:
            self.mps_regularization(0)
            exp_ham_list = list()
            for ham in ham_list:
                exp_ham_list.append(self.calculate_exp_ham(ham, self.tau/2))
            for exp_ham in exp_ham_list:
                self.evolute_all_once(exp_ham)
            exp_ham_list.reverse()
            for exp_ham in exp_ham_list:
                self.evolute_all_once(exp_ham)

            tmp_fidelity = self.calculate_fidelity(self.old_tensor_data)
            self.old_tensor_data = self.tensor_data.copy()
            tmp_error = (1 - tmp_fidelity)/self.n_site
            if tmp_error < (self.cut_off * self.bond_limit):
                print('error is too small, stop this iteration. 1 - fidelity = ' + str(tmp_error))
                tmp_error = 0
            tmp_error = tmp_error / self.tau
            n_iter = n_iter + 1
            bin_num = bin_num + 1
            if (tmp_error < self.tensor_acc) or (bin_num >= self.bin_max):
                self.tau = self.tau / 2
                bin_num = 0
                print('n_iter = ' + str(n_iter) + ' tau = ' + str(self.tau) + ' error = ' + str(tmp_error))
        self.mps_regularization(0)
        self.tensor_data[0] = self.tensor_data[0] / self.tensor_data[0].norm()

    def calculate_fidelity(self, other_mps):
        # do not use this if fidelity is too small
        nn = 0
        init_tensor = tc.einsum('abc,abd->cd', self.tensor_data[nn], other_mps[nn].conj())
        # init_norm = 0
        for nn in range(1, self.n_site):
            init_tensor = tc.einsum('cd,cef->def', init_tensor, self.tensor_data[nn])
            init_tensor = tc.einsum('def,deg->fg', init_tensor, other_mps[nn].conj())
            # init_norm = init_norm + tc.log(init_tensor.norm())
            # init_tensor = init_tensor / init_tensor.norm()
            # print(init_tensor.norm())
        return init_tensor.norm()

    def calculate_fidelity_log(self, other_mps):
        nn = 0
        init_tensor = tc.einsum('abc,abd->cd', self.tensor_data[nn], other_mps[nn].conj())
        init_norm = 0
        for nn in range(1, self.n_site):
            init_tensor = tc.einsum('cd,cef->def', init_tensor, self.tensor_data[nn])
            init_tensor = tc.einsum('def,deg->fg', init_tensor, other_mps[nn].conj())
            init_norm = init_norm + tc.log(init_tensor.norm())
            init_tensor = init_tensor / init_tensor.norm()
            # print(init_tensor.norm())
        return init_norm

    def calculate_energy(self, ham_list):

        energy = 0
        self.mps_regularization(0)
        self.tensor_data[0] = self.tensor_data[0] / self.tensor_data[0].norm()

        if not isinstance(ham_list, list):
            raise ValueError('ham needs to be a list')

        new_ham_list = ham_list.copy()

        while len(new_ham_list) > 0:
            sum_ham, n_interval_list = self.merge_ham(new_ham_list)
            interval_max = max(n_interval_list)
            while (self.regular_center + interval_max) <= self.n_site:
                tmp_energy = self.calculate_energy_once(sum_ham)
                energy = energy + tmp_energy
                if self.regular_center < (self.n_site - 1):
                    self.move_regular_center2next()
                else:
                    break

            new_ham_list = []
            for nn in range(len(n_interval_list)):
                if n_interval_list[nn] < interval_max:
                    new_ham_list.append(ham_list[nn])
        return energy

    def merge_ham(self, ham_list):
        n_interval_list = list()
        for ham in ham_list:
            n_interval_list.append(round(np.log2(ham.numel()) / 2))
        interval_max = max(n_interval_list)
        sum_ham = 0
        for nn in range(len(ham_list)):
            inter_diff = interval_max - n_interval_list[nn]
            one_matrix = tc.eye(2 ** inter_diff, device=self.device, dtype=self.dtype)
            tmp_ham = tc.kron(ham_list[nn], one_matrix)
            sum_ham = sum_ham + tmp_ham

        return sum_ham, n_interval_list

    def calculate_energy_once(self, ham):
        tmp_index = self.regular_center
        n_interval = round(np.log2(ham.numel()) / 2)
        if n_interval == 1:
            init_tensor = tc.einsum('abc,adc->bd', self.tensor_data[tmp_index], self.tensor_data[tmp_index].conj())
        else:
            init_tensor = tc.einsum('abc,ade->bdce', self.tensor_data[tmp_index], self.tensor_data[tmp_index].conj())
            end_tensor = tc.einsum('abc,xdc->axbd',
                                   self.tensor_data[tmp_index + n_interval - 1],
                                   self.tensor_data[tmp_index + n_interval - 1].conj())
            tmp_dv = 2
            for nn in range(1, n_interval - 1):
                tmp_dv = tmp_dv * 2
                tmp_shape = self.tensor_data[tmp_index + nn].shape
                init_tensor = tc.einsum('bdce,cfg,emn->bfdmgn', init_tensor,
                                        self.tensor_data[tmp_index + nn], self.tensor_data[tmp_index + nn].conj())

                init_tensor = init_tensor.reshape(tmp_dv, tmp_dv, tmp_shape[2], tmp_shape[2])
            tmp_dv = tmp_dv * 2
            init_tensor = tc.einsum('dbce,cefg->dfbg', init_tensor, end_tensor).reshape(tmp_dv, tmp_dv)
        energy = tc.einsum('ab,ab->', init_tensor, ham)
        if tc.abs(energy.imag) > 1e-15:
            raise ValueError('energy is complex')
        return energy.real









# Copyright (C) <2022>  <Zheng-Zhi Sun>
# This file is part of QuantumIntelligence.
# QuantumIntelligence is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Foobar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Foobar. If not, see <https://www.gnu.org/licenses/>.

from QuantumIntelligence.BasicFunSZZ import BasicClass


class TensorNetwork:

    def __init__(self):
        self.tensor_data = list()
        self.tensor_info = BasicClass.MultiDimensionDict()
        self.tmp = BasicClass.MultiDimensionDict()









# Copyright (C) <2022>  <Zheng-Zhi Sun>
# This file is part of QuantumIntelligence.
# QuantumIntelligence is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Foobar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Foobar. If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import pynvml
import os
import pickle
import torch as tc
from mpl_toolkits.mplot3d import Axes3D
import matplotlib.pyplot as plt


def info_contact():
    """Return the information of the contact"""
    info = dict()
    info['name'] = 'Sun Zheng-Zhi'
    info['email'] = 'sunzhengzhi.work@gmail.com'
    info['affiliation'] = 'Tsinghua University'
    return info

# These are from the original functions of Sun Zheng-Zhi


def get_best_gpu(device='cuda'):
    if isinstance(device, tc.device):
        return device
    elif device == 'cuda':
        pynvml.nvmlInit()
        num_gpu = pynvml.nvmlDeviceGetCount()
        memory_gpu = tc.zeros(num_gpu)
        for index in range(num_gpu):
            handle = pynvml.nvmlDeviceGetHandleByIndex(index)
            memory_info = pynvml.nvmlDeviceGetMemoryInfo(handle)
            memory_gpu[index] = memory_info.free
        max_gpu = int(tc.sort(memory_gpu, )[1][-1])
        return tc.device('cuda:' + str(max_gpu))
    elif device == 'cpu':
        return tc.device('cpu')
    else:
        return tc.device(device)


def sort_dict(a):
    b = dict()
    dict_index = sorted(a.keys())
    for index in dict_index:
        b[index] = a[index]
    return b


def is_subdict(a, b):
    pointer = True
    # check keys
    for key in a.keys():
        if not a.get(key) == b.get(key):
            pointer = False
            break
    return pointer


def save_pr_add_data(path, file, data, names):
    mkdir(path)
    if os.path.isfile(path+file):
        tmp = load_pr(path+file)
    else:
        tmp = {}
    s = open(path + file, 'wb')
    for ii in range(0, len(names)):
        tmp[names[ii]] = data[ii]
    pickle.dump(tmp, s)
    s.close()


def save_pr_del_data(path, file, names):
    mkdir(path)
    if os.path.isfile(path+file):
        tmp = load_pr(path+file)
    else:
        tmp = {}
    s = open(path + file, 'wb')
    for ii in range(0, len(names)):
        tmp.pop(names[ii])
    pickle.dump(tmp, s)
    s.close()


# These are from BasicFunctions of Ran Shi_ju with some changes


def save_pr(path, file, data, names):
    """
    Save the data as a dict in a file located in the path
    :param path: the location of the saved file
    :param file: the name of the file
    :param data: the data to be saved
    :param names: the names of the data
    Notes: 1. Conventionally, use the suffix \'.pr\'. 2. If the folder does not exist, system will
    automatically create one. 3. use \'load_pr\' to load a .pr file
    Example:
    >>> x = 1
    >>> y = 'good'
    >>> save_pr('/test', 'ok.pr', [x, y], ['name1', 'name2'])
      You have a file '/test/ok.pr'
    >>> z = load_pr('/test/ok.pr')
      z = {'name1': 1, 'name2': 'good'}
      type(z) is dict
    """
    try:
        mkdir(path)
    except FileExistsError:
        pass
    # print(os.path.join(path, file))
    s = open(path+file, 'wb')
    tmp = dict()
    for i in range(0, len(names)):
        tmp[names[i]] = data[i]
    pickle.dump(tmp, s)
    s.close()


def load_pr(path_file, names=None):
    """
    Load the file saved by save_pr as a dict from path
    :param path_file: the path and name of the file
    :param names: the specific names of the data you want to load
    :return  the file you loaded
    Notes: the file you load should be a  \'.pr\' file.
    Example:
        >>> x = 1
        >>> y = 'good'
        >>> z = [1, 2, 3]
        >>> save_pr('.\\test', 'ok.pr', [x, y, z], ['name1', 'name2', 'name3'])
        >>> A = load_pr('.\\test\\ok.pr')
          A = {'name1': 1, 'name2': 'good'}
        >>> y, z = load_pr('\\test\\ok.pr', ['y', 'z'])
          y = 'good'
          z = [1, 2, 3]
    """
    if os.path.isfile(path_file):
        s = open(path_file, 'rb')
        if names is None:
            data = pickle.load(s)
            s.close()
            return data
        else:
            tmp = pickle.load(s)
            if type(names) is str:
                data = tmp[names]
                s.close()
                return data
            elif (type(names) is list) or (type(names) is tuple):
                nn = len(names)
                data = list(range(0, nn))
                for i in range(0, nn):
                    data[i] = tmp[names[i]]
                s.close()
                return tuple(data)
    else:
        return None


def mkdir(path):
    """
       Create a folder at your path
       :param path: the path of the folder you wish to create
       :return: the path of folder being created
       Notes: if the folder already exist, it will not create a new one.
    """
    try:
        os.makedirs(path)
    finally:
        pass

# end of Ran Shi-Ju's code


def easy_plot(x, *y):

    plt.figure(figsize=(8, 4))
    # plt.plot(np.array(x), np.array(y), 'b*')
    for nn in range(len(y)):
        # plt.plot(np.array(x), np.array(y[nn]), '*', markersize=1)
        plt.plot(np.array(x), np.array(y[nn]), label=nn)
    if len(y) == 0:
        tmp_x = np.arange(len(x))
        plt.plot(tmp_x, np.array(x), label=0)
    plt.xlabel('x')
    plt.ylabel('y')
    plt.ylim()
    plt.title('default')
    plt.legend()
    plt.show()
    plt.close()


def easy_plot_3d(x, y, z):

    x, y = np.meshgrid(np.array(x), np.array(y))
    fig = plt.figure()
    ax = Axes3D(fig)
    ax.plot_surface(x, y, np.array(z).T, rstride=1, cstride=1, cmap='rainbow')
    ax.set_xlabel('x')
    ax.set_ylabel('y')
    ax.set_zlabel('z')
    plt.show()
    return fig


def compare_dict_1key(dict_list, key_com):
    pointer = True
    value = list()
    try:
        value.append(dict_list[0][key_com])
    except KeyError:
        pass
    for dict_one in dict_list:
        try:
            if dict_one[key_com] not in value:
                value.append(dict_one[key_com])
                pointer = False
        except KeyError:
            pointer = False
    return pointer, value


def seek_unique_value(a_list):
    check = []
    for xx in a_list:
        if xx not in a_list:
            check.append(xx)
    return check


if __name__ == '__main__':
    print(info_contact())



# Copyright (C) <2022>  <Zheng-Zhi Sun>
# This file is part of QuantumIntelligence.
# QuantumIntelligence is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Foobar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Foobar. If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from QuantumIntelligence.BasicFunSZZ import wheel_function as wf

sx = np.array([[0, 1], [1, 0]])
sy = np.array([[0, -1j], [1j, 0]])
sz = np.array([[1, 0], [0, -1]])


# [1, 0] is |0>, [0, 1] is |1>


def generate_hamiltonian(n_site=5, name='Heisenberg', para=None, periodic=False):

    i_2d = np.eye(2)
    i_2d_list = list(range(n_site))
    i_2d_list[0] = 1

    sx2 = np.kron(sx, sx)
    sy2 = np.kron(sy, sy)
    sz2 = np.kron(sz, sz)

    for nn in range(1, n_site):
        i_2d_list[nn] = np.kron(i_2d_list[nn - 1], i_2d)

    # give Hamiltonian
    h_matrix = np.zeros((2 ** n_site, 2 ** n_site))

    if name == 'Heisenberg':
        # H = sum(-s_i*s_j)
        if para is None:
            jx = -1
            jy = -1
            jz = -1
            hz = -1
            hx = 0
        else:
            jx = para['jx']
            jy = para['jy']
            jz = para['jz']
            hz = para['hz']
            hx = para['hx']
        # print(jx,jy,jz,hz,para)
        # add sz
        for nn in range(n_site):
            h_matrix = h_matrix - np.kron(np.kron(i_2d_list[nn], sz * hz), i_2d_list[n_site - 1 - nn])
        # add sx
        for nn in range(n_site):
            h_matrix = h_matrix - np.kron(np.kron(i_2d_list[nn], sx * hx), i_2d_list[n_site - 1 - nn])
        # add sx2
        for nn in range(n_site - 1):
            h_matrix = h_matrix - np.kron(np.kron(i_2d_list[nn], jx * sx2), i_2d_list[n_site - 2 - nn])
        # add sy2
        for nn in range(n_site - 1):
            h_matrix = h_matrix - np.kron(np.kron(i_2d_list[nn], jy * sy2), i_2d_list[n_site - 2 - nn])
        # add sz2
        for nn in range(n_site - 1):
            h_matrix = h_matrix - np.kron(np.kron(i_2d_list[nn], jz * sz2), i_2d_list[n_site - 2 - nn])

        # periodic
        if periodic:
            h_matrix = h_matrix - jx * np.kron(np.kron(sx, i_2d_list[n_site - 2]), sx)
            h_matrix = h_matrix - jy * np.kron(np.kron(sy, i_2d_list[n_site - 2]), sy)
            h_matrix = h_matrix - jz * np.kron(np.kron(sz, i_2d_list[n_site - 2]), sz)

    if name == 'MaxCut':
        if para is None:
            para = ((0, 1), (1, 2), (2, 3), (1, 4))
        for oo in para:
            # print(oo)
            h_matrix = h_matrix + wf.np_kron(
                i_2d_list[oo[0]], sz, i_2d_list[oo[1] - oo[0] - 1], sz, i_2d_list[n_site - oo[1] - 1])

    return h_matrix


def str2matrix(str_input):
    tmp_matrix = np.ones(1, 1)
    for xx in str_input:
        tmp_matrix = np.kron(tmp_matrix, str2matrix(xx))
    return tmp_matrix


def str2pauli(str_single):
    if str_single == 'x':
        return sx
    elif str_single == 'y':
        return sy
    elif str_single == 'z':
        return sz
    else:
        raise ValueError('Input string should be x, y, or z.')